/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2014 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.annotation.valueprovider;

import static org.fortiss.tooling.kernel.utils.EcoreUtils.pickFirstInstanceOf;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EEnum;
import org.eclipse.emf.ecore.EObject;
import org.fortiss.tooling.base.model.element.IAnnotatedSpecification;
import org.fortiss.tooling.base.model.element.IModelElement;

/**
 * Base class for {@link IAnnotationValueProvider}s.
 * 
 * @author diewald, barner
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating YELLOW Hash: 8D49FAC7EC65BC25421AC7E0DA0CAF01
 */
public abstract class ValueProviderBase<T extends IAnnotatedSpecification> implements
		IAnnotationValueProvider<T> {

	/**
	 * {@link EClass} of {@link IAnnotatedSpecification} that contains the respective structural
	 * feature and that is managed by this {@link IAnnotationValueProvider}.
	 */
	protected final EClass annotatedSpecificationEClass;

	/**
	 * TODO(VA) Probably obvious for somebody who knows already how to deal with annotations but not
	 * for a newcomer, so take this remark into account only if you think it's valuable: what is an
	 * instanceKey?
	 * 
	 * Map: instanceKey -> Fixed input choice to be offered in a combo box. The semantics of the map
	 * value is as follows:
	 * <ul>
	 * <li> {@code fixedInputChoice[key] == null} -> do not use fixed values, provide text input
	 * cell.</li>
	 * <li> {@code !fixedInputChoice[key].isEmpty()} -> provide a fixed set of choices in an input
	 * box.
	 * <li> {@code fixedInputChoice[key].isEmpty()} -> dynamically manage the set of provided values,
	 * i.e., the combo box is editable and enables the user to add new values and remove existing
	 * choices (by removing all instances of them).</li>
	 * <li>For enum types, the set of inputs is automatically initialized from the underlying
	 * {@link EEnum} type definition.</li>
	 * </ul>
	 */
	private Map<String, List<String>> fixedInputChoiceMap;

	/**
	 * Map: instanceKey -> Current input choice, computed according to the initialization of
	 * {@link #fixedInputChoiceMap}.
	 */
	private Map<String, Collection<String>> currentInputChoiceMap;

	/**
	 * Constructs a new {@link IAnnotationValueProvider} for an annotation specified in a given
	 * {@link EClass}.
	 */
	public ValueProviderBase(EClass annotatedSpecificationEClass) {
		this.annotatedSpecificationEClass = annotatedSpecificationEClass;

		fixedInputChoiceMap = new HashMap<String, List<String>>();
		currentInputChoiceMap = new HashMap<String, Collection<String>>();
	}

	/** {@inheritDoc} */
	@Override
	public boolean setAnnotationName(String name, T specification) {
		// Renaming this annotation is not supported by default.
		return false;
	}

	/** {@inheritDoc} */
	@Override
	public boolean canEdit(T specification) {
		return canEdit(specification, DEFAULT_KEY);
	}

	/** {@inheritDoc} */
	@Override
	public <U> U getAnnotationValue(T specification, String instanceKey) throws Exception {
		return getAnnotationValue(specification);
	}

	/** {@inheritDoc} */
	@Override
	public void setAnnotationValue(String value, T specification) throws Exception {
		setAnnotationValue(value, specification, null);
	}

	/** {@inheritDoc} */
	@Override
	public <U> void setAnnotationValue(U value, T specification) throws Exception {
		setAnnotationValue(value, specification, null);
	}

	/** {@inheritDoc} */
	@Override
	public void setAnnotationValue(String value, T specification, String instanceKey)
			throws Exception {
		throw new Exception(
				"setAnnotationValue(String) is not supported / has not been implemented for annotations of type " +
						specification.getClass());
	}

	/** {@inheritDoc} */
	@Override
	public <U> void setAnnotationValue(U value, T specification, String instanceKey)
			throws Exception {
		throw new Exception(
				"setAnnotationValue(U) is not supported / has not been implemented for annotations of type " +
						specification.getClass());
	}

	/** {@inheritDoc} */
	@Override
	public List<Class<? extends EObject>> excludeModelElementTypeFromAnnotatedSpecification() {
		return null;
	}

	/**
	 * Add decorations to an annotation when its instance is newly created by
	 * {@link #getAnnotatedSpecificationForModelElement(IModelElement)}.
	 * 
	 * @param specification
	 *            annotation to be decorated
	 */
	protected void decorateAnnotationSpecification(T specification) throws Exception {
		decorateAnnotationSpecification(specification, null);
	}

	/**
	 * Add decorations to an annotation when its instance is newly created by
	 * {@link #getAnnotatedSpecificationForModelElement(IModelElement)}.
	 * 
	 * @param specification
	 *            annotation to be decorated
	 * @param instanceKey
	 *            instanceKey of annotation to be decorated
	 */
	protected void decorateAnnotationSpecification(T specification, String instanceKey)
			throws Exception {
		// No decoration
	}

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	@Override
	public T getAnnotatedSpecificationForModelElement(IModelElement element) throws Exception {

		// TODO(VA,B11)
		T rval =
				(T)pickFirstInstanceOf(
						(Class<IAnnotatedSpecification>)annotatedSpecificationEClass
								.getInstanceClass(),
						element.getSpecifications());

		if(rval != null) {
			return rval;
		}

		// TODO(VA,B11)
		// Create annotation
		T specification =
				(T)annotatedSpecificationEClass.getEPackage().getEFactoryInstance()
						.create(annotatedSpecificationEClass);
		if(getInstanceKeys(specification).isEmpty()) {
			decorateAnnotationSpecification(specification);
		} else {
			for(String key : getInstanceKeys(specification)) {
				try {
					decorateAnnotationSpecification(specification, key);
				} catch(Exception e) {
					// Some of the keys might not be backed by an implementation that needs to
					// "decorate" the underlying specification object (e.g., a custom key that
					// represents a calculated, volatile value. Hence, just ignore invalid keys
					// during this setup phase. The other methods (get/set/...) will throw an
					// Exception on invalid keys.
				}
			}
		}

		// Hook specification to model element
		element.addSpecification(specification);

		return specification;
	}

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	@Override
	public Class<T> getAnnotationClazz() {
		return (Class<T>)annotatedSpecificationEClass.getInstanceClass();
	}

	/** {@inheritDoc} */
	@Override
	public Collection<String> getInstanceKeys(T specification) {
		return Collections.emptyList();
	}

	/** {@inheritDoc} */
	@Override
	public boolean allowsDynamicAnnotationInstances() {
		return false;
	}

	/** Returns the {@link #fixedInputChoiceMap}. */
	public Map<String, List<String>> getFixedInputChoiceMap() {
		return fixedInputChoiceMap;
	}

	/** Returns the {@link #currentInputChoiceMap}. */
	public Map<String, Collection<String>> getCurrentInputChoiceMap() {
		return currentInputChoiceMap;
	}

	/** Sets the {@link #currentInputChoiceMap}. */
	public void setCurrentInputChoiceMap(Map<String, Collection<String>> currentInputChoiceMap) {
		this.currentInputChoiceMap = currentInputChoiceMap;
	}

	/**
	 * Updates the current input choice, and returns the actual value. Called from
	 * {@link #setAnnotationValue(String, IAnnotatedSpecification)}. In particular, this
	 * method handles the removal of input choices from if these are dynamically managed.
	 * TODO(VA) "from" what?
	 * TODO(VA) What's actually the use of "value" here? First it's confusing to provide a
	 * value in a "get***" function. Second it's used only in the very first condition and in a very
	 * limited way. Is it provided as a facility to overloading methods? In any case, deserves an
	 * explanation in comment, I think.
	 * I must say overall I don't really get what this method is doing, but I'm not familiar with
	 * annotations.
	 */
	protected String getAnnotationValueAndUpdateInputChoice(String value, T specification,
			String instanceKey) throws Exception {
		// TODO(VA,B11)
		if((value == null || value.isEmpty()) && currentInputChoiceMap.get(instanceKey) != null &&
				!currentInputChoiceMap.get(instanceKey).isEmpty()) {
			// Only one entry left: do not remove
			if(currentInputChoiceMap.get(instanceKey).size() == 1) {
				return currentInputChoiceMap.get(instanceKey).iterator().next();
			}

			// 1st entry is to be removed: Choose 2nd entry as new value
			String currentValue = getAnnotationValue(specification);
			// TODO(VA) currentInputChoiceMap.get(instanceKey).iterator().next() could be factorized
			// in the next three lines
			if(currentInputChoiceMap.get(instanceKey).iterator().next().equals(currentValue)) {
				Iterator<String> iter = currentInputChoiceMap.get(instanceKey).iterator();
				iter.next();
				String rval = iter.next();
				currentInputChoiceMap.get(instanceKey).remove(currentValue);
				return rval;
			}

			// Otherwise, return the entry just before the entry to be removed.
			String rval = null;
			// TODO(VA) More concise:
			// for(String s : currentInputChoiceMap.get(instanceKey))
			// instead of the three following lines.
			for(Iterator<String> iter = currentInputChoiceMap.get(instanceKey).iterator(); iter
					.hasNext();) {
				String s = iter.next();
				if(s != null && s.equals(currentValue)) {
					currentInputChoiceMap.get(instanceKey).remove(currentValue);
					return rval;
				}
				rval = s;
			}
		}
		return null;
	}
}
